"""
A scraper that uses your username and password, in addition to your own app's client ID, client secret, and
redirect URI, to log in to Basecamp 3 as you and authorize your own app. This is how we can obtain an access token and
refresh token using your username and password. The request URI doesn't have to be a real URL but it does have to match
what you put as your app's redirect URI.
"""

from bs4 import BeautifulSoup
import requests
try:
    from urlparse import urljoin, urlparse, parse_qs
except ImportError:
    from urllib.parse import urljoin, urlparse, parse_qs


class TokenRequester(object):
    DOMAIN = "launchpad.37signals.com"

    # the authorization page for an app. (i.e. "Allow this app to access your data?")
    AUTH_URL = "https://{domain}/authorization/new?client_id={client_id}&redirect_uri={redirect_uri}&type=web_server"

    # pretend to be a browser
    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) " \
                 "Chrome/65.0.3325.181 Safari/537.36"

    def __init__(self, client_id, redirect_uri, user_email, user_pass, session=None):
        self.client_id = client_id
        self.redirect_uri = redirect_uri
        self.user_email = user_email
        self.user_pass = user_pass
        self._session = session

    def get_user_code(self):
        # check that we have the needed fields in our configuration to do this...
        errors = []
        if self.client_id is None:
            errors.append('client_id')
        if self.redirect_uri is None:
            errors.append('redirect_uri')
        if self.user_email is None:
            errors.append('user_email')
        if self.user_pass is None:
            errors.append('user_pass')
        if errors:
            raise ValueError("The following required fields to get an access token are missing: %s" % ", ".join(errors))

        if self._session is None:
            self._session = requests.session()

        # log in to Basecamp 3 using the username and password like a browser
        login_resp = self._use_login_form(self.client_id, self.redirect_uri, self.user_email, self.user_pass)
        # TODO do something if this login was unsuccessful

        # if successful, we're on the authorize this app page for the client ID provided
        user_code = self._submit_authorize_app_form(login_resp.url, login_resp.text)
        # TODO do something if this didn't get us a code

        return user_code

    def _use_login_form(self, client_id, redirect_uri, user_email, user_pass):
        # get the page
        auth_url = self.AUTH_URL.format(domain=self.DOMAIN, client_id=client_id, redirect_uri=redirect_uri)
        resp = self._session.get(auth_url)

        # use Beautiful Soup to find the login form and the fields on it
        soup = BeautifulSoup(resp.text, 'html.parser')
        login_form = soup.find('form', attrs={'data-behavior': 'login_form'})
        action_url = login_form.attrs['action']
        login_data = {
            'username': user_email,
            'password': user_pass,
            'utf8': "\u2713",  # checkmark
            'remember_me': 'true',
            'commit': 'Log in',
        }
        # authenticity_token appears to be a CSRF token on the form, not related to the Oauth2 tokens we're seeking
        login_data['authenticity_token'] = login_form.find('input', attrs={'name': 'authenticity_token'}).attrs['value']

        post_login_url = urljoin(auth_url, action_url)  # the URL to send the login form data to (form's action)
        login_resp = self._session.post(post_login_url, data=login_data)

        return login_resp

    def _submit_authorize_app_form(self, url, html):
        authorize_app_page = BeautifulSoup(html, 'html.parser')
        auth_form = authorize_app_page.find('form')
        authorize_data = {}
        for finput in auth_form.find_all('input'):
            authorize_data[finput.attrs['name']] = finput.attrs.get('value')
        auth_form_url = auth_form.attrs['action']
        authorize_post_url = urljoin(url, auth_form_url)
        app_token_resp = self._session.post(authorize_post_url, data=authorize_data, allow_redirects=False)

        # the redirect has a "code" on the end. This is what authorizes us to get the access_token
        app_redirect_uri = app_token_resp.headers['Location']
        # parse the code out of the redirect URI
        auth_code = parse_qs(urlparse(app_redirect_uri).query)['code'][0]

        return auth_code



import argparse
import logging
import os
import traceback

from basecampy3 import Basecamp3
from basecampy3.token_requestor import TokenRequester
from basecampy3.config import BasecampConfig


try:
   input = raw_input  # Python 2
except NameError:  # Python 3
    pass


class CLI(object):
    def __init__(self):
        pass

    @classmethod
    def from_command_line(cls):
        new = cls()
        parser = argparse.ArgumentParser("bc3", description="BasecamPY3 API Tool")
        parser.add_argument("--debug", "--verbose", dest="debug", action="store_true",
                            help="Enables more verbose output")
        # parser.add_argument('command', help="The section of the API to access.")
        subparsers = parser.add_subparsers(title="subcommands", description="valid subcommands")
        configure = subparsers.add_parser('configure', help="Configure tokens for this account")
        configure.set_defaults(func=cls._configure)
        # projects = subparsers.add_parser("projects", help="Manipulate project data in Basecamp")
        # version = subparsers.add_parser("version", help="Displays the installed version of BasecamPY3")
        args = parser.parse_args()
        loglevel = logging.DEBUG if args.debug else logging.INFO
        logging.getLogger().setLevel(loglevel)
        logging.basicConfig()
        try:
            args.func()
        except AttributeError:
            parser.print_usage()
            return

        return new

    @staticmethod
    def _configure():
        print("This will generate an access token and refresh token for using the Basecamp 3 API.")
        print("If you have not done so already, you need to create an app at:")
        print("https://launchpad.37signals.com/integrations")

        client_id = input("What is your app's Client ID?").strip()
        client_secret = input("What is your app's Client Secret?").strip()
        redirect_uri = input("What is your app's Redirect URI?").strip()
        print("For this next bit we need your username and password. When you have finished this wizard and have "
              "obtained your access and request tokens, you can change your password as the refresh token is all we "
              "will need.")
        user_email = input("What is the email address you log into Basecamp with?").strip()
        user_pass = input("What is your password for logging into Basecamp?").strip()
        print("Obtaining your access key and refresh token...")
        requestor = TokenRequester(client_id, redirect_uri, user_email, user_pass)
        code = requestor.get_user_code()
        tokens = Basecamp3.trade_user_code_for_access_token(client_id=client_id, redirect_uri=redirect_uri,
                                                            client_secret=client_secret, code=code)
        print("Success! Your tokens are listed below.")
        print("Access Token: %s" % tokens['access_token'])
        print("Refresh Token: %s" % tokens['refresh_token'])
        while True:
            should_save = input("Do you want to save? (Y/N)").upper().strip()
            if should_save in ("Y", "YES"):
                should_save = True
                break
            elif should_save in ("N", "NO"):
                should_save = False
                break
            else:
                print("Sorry I don't understand. Please enter Y or N.")
        if should_save is False:
            return
        while True:
            location = input("Where should I save? (default ~/.conf/basecamp.conf)")
            location = location.strip()
            if not location:
                location = os.path.expanduser("~/.conf/basecamp.conf")
            try:
                conf = BasecampConfig(client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri,
                                      access_token=tokens['access_token'], refresh_token=tokens['refresh_token'])
                conf.save(location)
                break
            except Exception as ex:
                logging.error(traceback.format_exc())
                print("Failed to save to '%s'" % location)

    # @staticmethod
    # def _get_modules():
    #     # find all endpoint modules
    #     endpoint_modules = inspect.getmembers(basecampy3.endpoints, inspect.ismodule)
    #     modules = []
    #     endpoint_map = {}
    #     for module_name, module in endpoint_modules:
    #         if module_name in ('_base', 'util'):  # exclude non-endpoint modules
    #             continue
    #         endpoint_dict = {}
    #
    #         classes = inspect.getmembers(module, inspect.isclass)
    #         for classname, cls in classes:
    #             # find the class for this Endpoint
    #             if cls is not BasecampEndpoint and issubclass(cls, BasecampEndpoint):
    #                 # this is our Endpoint, get its non-private methods
    #                 methods_dict = {mthd[0]: mthd for mthd in inspect.getmembers(cls, inspect.ismethod)
    #                                 if not mthd[0].startswith('_')}
    #
    #                 endpoint_dict[classname] = methods_dict
    #
    #                 endpoint_map[module_name] = endpoint_map
    #                 break
    #
    #         issubclass(inspect.getmembers(module, inspect.isclass)[3][1], basecampy3.endpoints._base.BasecampEndpoint)
    #     return {m[0]: m[1] for m in endpoint_modules if m[0] not in ("_base", "util")}


def main():
    return CLI.from_command_line()


if __name__ == "__main__":
    cli = main()



from ._base import BasecampObject, BasecampEndpoint
from ..constants import DOCK_NAME_CAMPFIRE
from . import util


class Campfire(BasecampObject):
    def __int__(self):
        return int(self.id)

    @property
    def lines(self):
        """
        Get the messages posted in this Campfire. Messages are known as "lines" so they don't get confused with
        Campfire's Message Board.
        :return: the CampfireLine objects in this Campfire
        :rtype: collections.Iterable[basecampy3.endpoints.campfire_lines.CampfireLine]
        """
        return self._endpoint._api.campfire_lines.list(campfire=self)

    @property
    def project(self):
        """
        Get the Project object this Campfire belongs to
        :return: the Project object this Campfire belongs to
        """
        if self.bucket['type'] == 'Project':
            return self._endpoint._api.projects.get(self.bucket['id'])
        return None

    def post_message(self, content):
        """
        Post a new line (message) in this campfire from the current user with the provided `content` string.
        Can be formatted in HTML.

        :param content: message string that can be HTML-formatted
        :type content: str
        :return: the new CampfireLine object
        :rtype: basecampy3.endpoints.campfire_lines.CampfireLine
        """
        return self._endpoint._api.campfire_lines.create(content, campfire=self)

    def __str__(self):
        return "Campfire %s: '%s'" % (self.id, self.bucket['name'])


class Campfires(BasecampEndpoint):
    OBJECT_CLASS = Campfire

    GET_URL = "{base_url}/buckets/{project_id}/chats/{campfire_id}.json"
    LIST_URL = "{base_url}/chats.json"

    def get(self, project=None, campfire=None):
        """
        Get just one campfire by its ID and a Project's ID. Provide a Project object or Campfire to skip
        the other parameter (i.e. if you pass a Campfire object, you do not need to specify the project parameter)

        :param campfire: the Campfire or Campfire's ID
        :param project: the Project object or Project's ID

        :return: a Campfire object if found, raises an error if not found
        """
        project_id, campfire_id = util.project_or_object(project, campfire, section_name=DOCK_NAME_CAMPFIRE)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, campfire_id=campfire_id)
        return self._get(url)

    def list(self):
        """
        Get a list of all active Campfires visible to the current user.

        :return: a list of Campfire objects
        """
        url = self.LIST_URL.format(base_url=self.url)
        return self._get_list(url)




"""
A configuration object used by the API object. Looks for a basecamp.conf file at startup and makes it a variable
called GLOBAL. If nothing is specified when you make a Basecamp API object, the GLOBAL configuration will be used.

Reads from INI files. Also writes back to the INI files if you use this API to obtain access or refresh tokens.
"""
try:
    from ConfigParser import SafeConfigParser as ConfigParser, NoSectionError, NoOptionError
except ImportError:
    from configparser import ConfigParser, NoSectionError, NoOptionError

from datetime import datetime, timedelta
import logging
import os
import six

from .exc import *


class BasecampConfig(object):
    DEFAULT_CONFIG_FILE_LOCATIONS = [
        "basecamp.conf",  # current directory
        os.path.expanduser("~/.conf/basecamp.conf"),  # user profile directory/.conf/basecamp.conf
        "/etc/basecamp.conf",
    ]

    def __init__(self, client_id=None, client_secret=None, redirect_uri=None, user_email=None, user_pass=None,
                 access_token=None, access_expires=None, refresh_token=None, filepath=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.user_email = user_email
        self.user_pass = user_pass
        self.access_token = access_token
        self._access_expires = datetime.min
        if access_expires is not None:
            self.access_expires = access_expires
        self.refresh_token = refresh_token
        self.filepath = filepath

    @property
    def access_expires(self):
        return self._access_expires

    @access_expires.setter
    def access_expires(self, value):
        if isinstance(value, datetime):
            self._access_expires = value
            return

        try:
            value = float(value)
            value = datetime.now() + timedelta(seconds=value)
        except:
            raise ValueError("`access_expires` needs to be a float (number of seconds until expiration) "
                             "or a datetime. Got a `%s`." % type(value).__name__)
        self._access_expires = value

    def read(self, filepath):
        config = ConfigParser()
        config.read(filepath)
        attrs = [k for k in self.__dict__.keys() if k not in ('filepath', '_access_expires')]
        attrs.append('access_expires')
        for key in attrs:
            try:
                value = config.get('BASECAMP', key)
                if key == "access_expires":
                    value = datetime.utcfromtimestamp(float(value))
                setattr(self, key, value)
            except NoOptionError:
                pass
        self.filepath = filepath

    def save(self, filepath=None):
        if filepath is None:
            filepath = self.filepath
        try:
            os.makedirs(os.path.dirname(filepath), mode=0o770)
        except OSError:
            pass  # folder probably already exists

        config = ConfigParser()
        config.add_section('BASECAMP')
        attrs = [k for k in self.__dict__.keys() if k != 'filepath']
        for key in attrs:
            try:
                value = getattr(self, key)
                if value is None:
                    continue
                if key == "_access_expires":
                    key = "access_expires"
                    try:
                        value = value.timestamp()
                    except:
                        continue  # if timestamp is at min value, this will fail, that's ok

                config.set('BASECAMP', key, six.text_type(value))
                setattr(self, key, value)
            except (NoSectionError, NoOptionError):
                pass
        with open(filepath, "w") as fileout:
            config.write(fileout)
        self.filepath = filepath

    @classmethod
    def from_filepath(cls, filepath):
        if not os.path.exists(filepath):
            raise IOError("Non-existent configuration file '%s'" % filepath)
        new_config = cls()
        new_config.read(filepath)
        return new_config

    @classmethod
    def load_from_default_paths(cls):
        for config_file in cls.DEFAULT_CONFIG_FILE_LOCATIONS:
            try:
                return cls.from_filepath(config_file)
            except Exception as ex:
                logging.debug("%s: %s is missing or invalid.", type(ex).__name__, config_file)
        else:
            raise NoDefaultConfigurationFound()



from ..constants import DOCK_NAME_MESSAGE_BOARD
from . import recordings, util


class Message(recordings.Recording):
    """
    A Message that was posted on a Message Board. Not to be confused with a Campfire Line.
    """

    def edit(self, subject=False, content=False, category=False):
        """
        Change the subject, content, and/or category of this message.
        False is a placeholder for no-change. None erases the value like a blank string.

        :param subject: change the subject (title) of the message.
        :param content: change the body of the message
        :param category: change the category of the message
        """

        self._endpoint.update(subject=subject, content=content, category=category, message=self)

    @property
    def icon(self):
        try:
            return self.category['icon']
        except AttributeError:
            return ""

    @property
    def message_board(self):

        board_id = self.parent['id']
        project_id = self.bucket['id']
        return self._endpoint._api.message_boards.get(project=project_id, board=board_id)

    def __int__(self):
        return int(self.id)

    def __str__(self):
        try:
            text = "Message {id} Created by {creator} <{email}>: {icon} '{subject}'".format(
                id=self.id, creator=self.creator['name'], email=self.creator['email_address'], icon=self.icon,
                subject=self.subject
            )
            return text
        except:
            return super(Message, self).__str__()


class Messages(recordings.RecordingEndpoint):
    OBJECT_CLASS = Message

    CREATE_URL = "{base_url}/buckets/{project_id}/message_boards/{board_id}/messages.json"
    GET_URL = "{base_url}/buckets/{project_id}/messages/{message_id}.json"
    LIST_URL = "{base_url}/buckets/{project_id}/message_boards/{board_id}/messages.json"
    UPDATE_URL = "{base_url}/buckets/{project_id}/messages/{message_id}.json"

    def create(self, subject, content=None, status="active", category=None, project=None, board=None):
        """
        Create a new Message object. Either a project ID and MessageBoard ID must be given or just a MessageBoard
        object.

        :param subject: the subject (title) of this message
        :type subject: str
        :param content: the content (body) of this message. Can be HTML.
        :type content: str
        :param status: the status of this message. Set to "active" to post instantly. Required.
        :type status: str
        :param category: what category this message should be. MessageCategory objects vary by project.
        :type category: basecampy3.endpoints.message_categories.MessageCategory|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param board: the MessageBoard to post this message to. Either specify the object or the ID.
        :type board: basecampy3.endpoints.message_boards.MessageBoard|int
        :return: the newly posted Message
        :rtype: Message
        """
        project_id, board_id = util.project_or_object(project, board, section_name=DOCK_NAME_MESSAGE_BOARD)
        data = {
            "subject": subject,
            "status": status,
        }
        if content is not None:
            data["content"] = content
        if category is not None:
            data["category_id"] = int(category)
        url = self.CREATE_URL.format(base_url=self.url, project_id=project_id, board_id=board_id)
        message = self._create(url, data=data)
        return message

    def get(self, message, project=None):
        """
        Get a Message by using a Message object or a message ID with project ID.

        :param message: a Message object or ID
        :type message: Message|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :return: a Message object if found, raises an error if not found
        :rtype: Message
        """
        project_id, message_id = util.project_or_object(project, message)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, message_id=message_id)
        return self._get(url)

    def list(self, project=None, board=None):
        """
        Get a list of Messages from a given MessageBoard.

        :return: a list of Message objects
        :rtype: collections.Iterable[Message]
        """
        project_id, board_id = util.project_or_object(project, board, section_name=DOCK_NAME_MESSAGE_BOARD)
        url = self.LIST_URL.format(base_url=self.url, project_id=project_id, board_id=board_id)
        return self._get_list(url)

    def update(self, subject=False, content=False, category=False, project=None, message=None):
        """
        Change the subject, content, and/or category of a given message.
        False is a placeholder for no-change. None erases the value like a blank string.

        :param subject: change the subject (title) of the message.
        :type subject: str
        :param content: change the body of the message
        :type content: str
        :param category: change the category of the message
        :type category: basecampy3.endpoints.message_categories.MessageCategory|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param message: a Message object or ID
        :type message: Message|int
        """
        if subject is False and content is False and category is False:
            raise ValueError("At least one of subject, content, and category parameters should have a value to "
                             "be changed.")
        project_id, message_id = util.project_or_object(project, message, section_name=DOCK_NAME_MESSAGE_BOARD)
        data = {}
        if subject is not False:
            data['subject'] = "" if subject is None else subject
        if content is not False:
            data['content'] = "" if content is None else content
        if category is not False:
            data['category_id'] = int(category)
        url = self.UPDATE_URL.format(base_url=self.url, project_id=project_id, message_id=message_id)
        self._update(url, data=data)



API_URL = "https://3.basecampapi.com/"

DOCK_NAME_CAMPFIRE = 'chat'
DOCK_NAME_MESSAGE_BOARD = 'message_board'
DOCK_NAME_TODOS = 'todoset'
DOCK_NAME_SCHEDULE = 'schedule'
DOCK_NAME_CHECKIN = 'questionnaire'
DOCK_NAME_VAULT = 'vault'
DOCK_NAME_FORWARDS = 'inbox'

RATE_LIMIT_REQUESTS = 50
RATE_LIMIT_PER_SECONDS = 10



from ..exc import *
from .. import constants

import re
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin


class BasecampObject(object):
    def __init__(self, json_dict, endpoint):
        """
        A Basecamp object retrieved from the API. The fields are stored in a special `_values` dictionary to
        accomodate future updates to the Basecamp 3 API. The fields inside the `_values` dictionary can be retrieved
        as if they were attributes of this object. Anything deeper than the first level object has to be done as a
        dictionary access.

        :param json_dict: a dictionary representing the parsed JSON returned by an API call
        :type json_dict: dict
        :param endpoint: a BasecampEndpoint for easy access to the Basecamp3 object that created this object
        :type endpoint: BasecampEndpoint
        """
        self._values = json_dict
        self._endpoint = endpoint

    def refresh(self, url=None):
        """
        Refresh this object's values from the API
        """
        if url is None:
            if 'url' in self._values:
                url = self._values['url']
            else:
                ex = ValueError("Can't refresh {object} without a URL".format(object=type(self).__name__))
                raise ex
        new_item = self._endpoint._get(url)  # luckily this object has the URL we can refresh from
        self._values.clear()
        self._values.update(new_item._values)

    def __getattr__(self, item):
        try:
            return self._values[item]
        except KeyError:
            # ex = "'{type}' object has no attribute '{item}'".format(type=self.__class__.__name__, item=item)
            ex = "'{type}' object has no attribute '{item}'".format(type='BasecampObject', item=item)
            raise AttributeError(ex)

    def __int__(self):
        return int(self.id)

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, repr(self._values))

    def __setattr__(self, key, value):
        if key in {"_values", "_endpoint"} or key not in self._values:
            return super(BasecampObject, self).__setattr__(key, value)
        else:
            self._values[key] = value

    def __str__(self):
        return "{type}".format(type=type(self).__name__)


class BasecampEndpoint(object):
    OBJECT_CLASS = BasecampObject
    URL = constants.API_URL
    _LINK_HEADER_URL_REGEX = re.compile(r'<(https.+)>')

    def __init__(self, api):
        """
        :param api: the API object that handles authorization and HTTP requests
        :type api: basecamp3.bc3_api.Basecamp3
        """
        self._api = api
        self.url = urljoin(self.URL, "/%s" % api.account_id)

    def _get_list(self, url, params=None, method="GET"):
        """
        Basecamp 3's API returns a paginated list of elements for most GET list endpoints. It has a geared pagination
        ratio so page 1 has 15 objects, page 2 has 30, page 3 has 50, and pages 4 and up have 100 objects each. This
        function returns a generator that can be looped through. The next page is transparently fetched when a user's
        loop has exhausted the current page.

        Someone could probably re-write this in a way that grabbed the next page asynchronously ahead of time...

        :param url: the URL to GET a list from
        :type url: str
        :param params: the GET parameters to add to the url after the "?" like "/projects.json?key=value&key2=value2"
        :type params: dict
        :param method: the HTTP verb to use when fetching this URL. Usually "GET".
        :type method: str
        :return: a generator that produces the requested objects
        :rtype: collections.Iterable[BasecampObject]
        """
        request_args = {'method': method, 'url': url}
        if params is not None:
            request_args['params'] = params

        return self._paginated_generator(request_args)

    def _get(self, url, method="GET"):
        resp = self._api._session.request(method, url)
        if not resp.ok:
            raise Basecamp3Error(response=resp)
        item = resp.json()
        return self.OBJECT_CLASS(item, self)

    def _create(self, url, data, method="POST", object_class=None):
        resp = self._api._session.request(method, url, json=data)
        if not resp.ok:
            raise Basecamp3Error(response=resp)
        json_data = resp.json()
        if object_class is None:
            object_class = self.OBJECT_CLASS
        item = object_class(json_data, self)
        return item

    def _update(self, url, data, method="PUT"):
        resp = self._api._session.request(method, url, json=data)
        if not resp.ok:
            raise Basecamp3Error(response=resp)
        json_data = resp.json()
        item = self.OBJECT_CLASS(json_data, self)
        return item

    def _no_response(self, url, data=None, method="PUT"):
        request_args = {"url": url, "method": method}
        if data is not None:
            request_args['json'] = data
        resp = self._api._session.request(**request_args)
        if not resp.ok:
            raise Basecamp3Error(response=resp)

    def _delete(self, url, method="DELETE"):
        resp = self._api._session.request(method, url)
        if not resp.ok:
            raise Basecamp3Error(response=resp)
        return resp

    def _paginated_generator(self, request_args):
        """
        Automatically gets the next page when getting paginated results, yielding each object on each page.

        :param request_args: kwargs for Session.request method
        :type request_args: dict
        """
        while request_args:
            resp = self._api._session.request(**request_args)
            if not resp.ok:
                raise Basecamp3Error(response=resp)
            link_header = resp.headers.get("Link")
            if link_header:
                next_page_url = self._LINK_HEADER_URL_REGEX.findall(link_header)[0]
                request_args = {'url': next_page_url, 'method': 'GET'}  # get ready to call the next page
            else:
                request_args = None  # clear it so we break the loop
            items_json = resp.json()
            for jdict in items_json:
                item = self.OBJECT_CLASS(jdict, self)  # convert JSON dict into a BasecampObject
                yield item



from . import _base, projects, util
from .. import constants


class TodoSet(_base.BasecampObject):
    def create(self, name, description=None):
        """
        Create a new TodoList object in this Project.

        :param name: the name for this TodoList
        :param description: (optional) an HTML-formatted description of this TodoList
        :return: the newly created TodoList
        :rtype: basecampy3.endpoints.todolists.TodoList
        """
        return self._endpoint._api.todolists.create(name=name, description=description, todoset=self)

    def list(self):
        """
        :return: a list of the TodoList objects in this Project
        :rtype: collections.Iterable[basecampy3.endpoints.todolists.TodoList]
        """
        return self._endpoint._api.todolists.list(todoset=self)


class TodoSets(_base.BasecampEndpoint):
    OBJECT_CLASS = TodoSet

    GET_URL = "{base_url}/buckets/{project_id}/todosets/{todoset_id}.json"

    def get(self, project, todoset=None):
        """
        Get a TodoSet object either from a project ID and a todoset ID or just a Project object.

        :param project: a Project object or ID
        :type project: projects.Project|int
        :param todoset: a TodoSet object or ID
        :type todoset: TodoSet|int
        :return: a TodoSet object
        :rtype: TodoSet
        """
        project_id, todoset_id = util.project_or_object(project, todoset, section_name=constants.DOCK_NAME_TODOS)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, todoset_id=todoset_id)
        return self._get(url)



"""
The only object you should have to import. Access all of the API from here.
"""
from datetime import datetime
import dateutil.parser
import logging
import requests
from .transport_adapter import Basecamp3TransportAdapter
import traceback

from basecampy3.exc import *
from .config import BasecampConfig
from .token_requestor import TokenRequester
from .endpoints import campfires
from .endpoints import campfire_lines
from .endpoints import messages
from .endpoints import message_boards
from .endpoints import message_categories
from .endpoints import people
from .endpoints import projects
from .endpoints import project_constructions
from .endpoints import templates
from .endpoints import todolists
from .endpoints import todolist_groups
from .endpoints import todos
from .endpoints import todosets


class Basecamp3(object):
    OAUTH_DOMAIN = "launchpad.37signals.com"
    ACCESS_TOKEN_URL = "https://{domain}/authorization/token?type=web_server&client_id={client_id}&" \
                       "redirect_uri={redirect_uri}&client_secret={client_secret}&code={code}"
    REFRESH_TOKEN_URL = "https://{domain}/authorization/token?type=refresh&refresh_token={0.refresh_token}&" \
                        "client_id={0.client_id}&redirect_uri={0.redirect_uri}&client_secret={0.client_secret}"
    USER_AGENT = "My Test Python API (phillip@4stromberg.com)"

    def __init__(self, client_id=None, client_secret=None, redirect_uri=None, access_token=None, access_expires=None,
                 refresh_token=None, user_email=None, user_pass=None, conf=None):
        """
        Create a new Basecamp 3 API connection. The following combinations of parameters are valid:

        1. `access_token` only
            Can access the API for now, but when this `access_token` expires you'll have to provide a new one. Not
            ideal for automation.
        2. `refresh_token`
            With `client_id`, `client_secret`, `redirect_uri`, and a `refresh_token`, we can obtain and refresh our
            own `access_token`. `refresh_token`s don't seem to have a limit so this is ideal for automation.
        3. `user_email` and `user_pass`
            With `user_email`, `user_pass`, `client_id`, `client_secret`, and `redirect_uri`, we can obtain our own
            `refresh_token` and therefore our own `access_token`. This is ideal for first runs. If your configuration
            file is in a writable spot, your `access_token` and `refresh_token` can be written to it automatically.
            Once you have obtained a `refresh_token`, you should remove your `user_email` and `user_pass` from the
            configuration and then change your password on Basecamp's site as a best practice.

        `client_id`, `client_secret`, and `redirect_uri` can be obtained by creating a new app here:
            https://launchpad.37signals.com/integrations

        `redirect_uri` doesn't have to be a real URL, but it does have to match what you put when you registered your
        app at the site above.

        `access_token` and `refresh_token` are normally obtained when a user clicks your app integration page and
        presses the big green "Yes, I'll allow access" button, which redirects them to your `redirect_uri` with a
        authorization code attached to it. Your app then uses its client_secret and client_id to obtain the tokens with
        this authorization code. Since that is a long and confusing process that requires you to set up a web server
        to answer the `redirect_uri`, see below on how this API makes it easy to obtain.

        `user_email` and `user_pass` is a login for Basecamp. It is NOT recommended to use this long term. If your
        Basecamp configuration file is in a writeable location, this API will write the `access_token` and
        `refresh_token` to it. You can then delete your username and password from the configuration file. Refresh
        tokens have no limit. There is no reason to continue using your username and password once you have obtained it.

        :param client_id: your app's client_id is given to you when you create a new app
        :type client_id: str
        :param client_secret: your app's client_secret is given to you when you create a new app
        :type client_secret: str
        :param redirect_uri: your app's redirect_uri can be fake but it needs to match on your app page
        :type redirect_uri: str
        :param access_token: an access token you have obtained from a user
        :type access_token: str
        :param access_expires: (optional) when the access token will expire as a datetime or datetime.timestamp float
        :type access_expires: datetime.datetime|float
        :param refresh_token: a refresh token obtained from a user, used for obtaining a new access_token
        :type refresh_token: str
        :param user_email: a user's login email for Basecamp 3. Only used to obtain refresh_token.
        :type user_email: str
        :param user_pass: a user's login password for Basecamp 3. Only used to obtain refresh_token.
        :type user_pass: str
        :param conf: a BasecampConfig object with all the settings we need so that we don't have to fill out all
                         these parameters
        :type conf: basecampy3.config.BasecampConfig
        """
        can_request_our_own_refresh_tokens = client_id and client_secret and redirect_uri and user_email and user_pass
        can_refresh_access_tokens = refresh_token and client_id and client_secret and redirect_uri
        if access_token or can_refresh_access_tokens or can_request_our_own_refresh_tokens:
            conf = BasecampConfig(client_id=client_id, client_secret=client_secret, redirect_uri=redirect_uri,
                                  user_email=user_email, user_pass=user_pass, access_token=access_token,
                                  refresh_token=refresh_token, access_expires=access_expires)
        elif conf is None:
            conf = BasecampConfig.load_from_default_paths()
        else:
            raise ValueError("Need a valid BasecampConfig object, a refresh token, an access token, or enough "
                             "information to get our own refresh tokens (client_id, client_secret, redirect_uri, "
                             "username, and password)")

        self._conf = conf
        self._session = requests.session()
        self._session.mount("https://", adapter=Basecamp3TransportAdapter())
        self._session.headers['User-Agent'] = self.USER_AGENT
        self.account_id = None
        self._authorize()
        self.campfires = campfires.Campfires(self)
        self.campfire_lines = campfire_lines.CampfireLines(self)
        self.messages = messages.Messages(self)
        self.message_boards = message_boards.MessageBoards(self)
        self.message_categories = message_categories.MessageCategories(self)
        self.people = people.People(self)
        self.projects = projects.Projects(self)
        self.project_constructions = project_constructions.ProjectConstructions(self)
        self.templates = templates.Templates(self)
        self.todolists = todolists.TodoLists(self)
        self.todolist_groups = todolist_groups.TodoListGroups(self)
        self.todos = todos.Todos(self)
        self.todosets = todosets.TodoSets(self)

    @property
    def who_am_i(self):
        data = self._get_data("https://%s/authorization.json" % self.OAUTH_DOMAIN)
        return data.json()

    @classmethod
    def trade_user_code_for_access_token(cls, client_id, redirect_uri, client_secret, code, session=None):
        access_token_url = cls.ACCESS_TOKEN_URL.format(domain=cls.OAUTH_DOMAIN, client_id=client_id,
                                                       redirect_uri=redirect_uri, client_secret=client_secret,
                                                       code=code)
        if session is None:
            session = requests.session()
        resp = session.post(access_token_url)  # this should be a JSON response with our access and refresh tokens
        if not resp.ok:
            raise InvalidUserCodeError(response=resp)

        token_json = resp.json()  # the None values should be replaced with the JSON values now
        return token_json

    def _get_data(self, url, auto_reauthorize=True):
        times_to_try = 2 if auto_reauthorize else 1
        resp = None
        for attempt in range(0, times_to_try):
            resp = self._session.get(url)
            if resp.status_code == 401 and auto_reauthorize:
                if attempt == (times_to_try - 1):  # was our last attempt
                    break
                else:
                    self._authorize()  # reauthorize
                    continue
            break  # we were authorized so we can stop attempting

        code = resp.status_code

        if code == 401:
            raise UnauthorizedError(message="Unable to authorize ourselves to Basecamp.", response=resp)
        return resp

    def _authorize(self):
        """
        Determine if we have the credentials at our disposal to make API calls.
        :return:
        """

        if self._is_token_expired():
            self._get_access_token()
        self._apply_token_to_headers()  # set our access token to the Authorization header for this session
        self.account_id = self._get_account_id()  # set our Basecamp account ID used in many API calls

    def _apply_token_to_headers(self):
        self._session.headers['Authorization'] = 'Bearer %s' % self._conf.access_token

    def _get_access_token(self):
        if self._conf.refresh_token:
            try:
                token_json = self._refresh_access_token()
                self._save_token_json(token_json)
                return  # if there wasn't an error, the access token is ready to go
            except InvalidRefreshTokenError:
                self._conf.refresh_token = None  # this is a bad token

        token_req = TokenRequester(self._conf.client_id, self._conf.redirect_uri,
                                   self._conf.user_email, self._conf.user_pass)
        code = token_req.get_user_code()

        token_json = self.trade_user_code_for_access_token(client_id=self._conf.client_id,
                                                           redirect_uri=self._conf.redirect_uri,
                                                           client_secret=self._conf.client_secret, code=code)
        self._save_token_json(token_json)

    def _get_account_id(self):
        identity = self.who_am_i
        for acct in identity['accounts']:
            if acct['product'] == 'bc3':
                return acct['id']
        raise UnknownAccountIDError("Could not determine this Basecamp account's ID")

    def _is_token_expired(self):
        """
        Check the access token in our configuration and see if it's expired. If it is expired or there isn't an access
        token, return Trues
        :return: True if token is missing or expired, False if the token is still valid
        """
        if not self._conf.access_token:
            return True
        self._apply_token_to_headers()  # ok then lets apply this token to our session if it's not there already
        try:
            resp = self._get_data("https://%s/authorization.json" % self.OAUTH_DOMAIN, False)
        except Exception as ex:
            logging.debug("Token is expired: %s" % traceback.format_exc())
            return True
        expires_at = dateutil.parser.parse(resp.json()['expires_at'])
        expires_at = expires_at.replace(tzinfo=None)
        return datetime.utcnow() >= expires_at

    def _refresh_access_token(self):
        url = self.REFRESH_TOKEN_URL.format(self._conf, domain=self.OAUTH_DOMAIN)
        resp = self._session.post(url)
        if resp.status_code != 200:
            raise InvalidRefreshTokenError(response=resp)
        token_json = resp.json()

        return token_json

    def _save_token_json(self, token_json):
        if 'access_token' in token_json:
            self._conf.access_token = token_json['access_token']
        if 'refresh_token' in token_json:
            self._conf.refresh_token = token_json['refresh_token']
        if 'expires_in' in token_json:
            self._conf.access_expires = token_json['expires_in']
        self._conf.save()



from . import recordings, util
import six


class TodoItem(recordings.Recording):
    def check(self):
        """
        Mark this TodoItem as complete.
        """
        self._endpoint.complete(todoitem=self, project=self.project_id)

    def uncheck(self):
        """
        Mark this TodoItem as incomplete.
        """
        self._endpoint.uncomplete(todoitem=self, project=self.project_id)

    def reposition(self, position):
        """
        Change the position of this TodoItem in the TodoList. 1 will put it at the top of the list.

        :param position: the new position for this TodoItem in the list. Must be greater than or equal to 1.
        :type position: int
        """
        self._endpoint.reposition(position=position, todoitem=self, project=self.project_id)

    def save(self, notify=None):
        """
        Saves the fields in this TodoItem using the update method in the Todos endpoint.

        :param notify: whether or not to notify subscribers of the changes
        :type notify: bool
        """
        self._endpoint.update(todoitem=self, project=self.project_id, content=self.content,
                              description=self.description, assignee_ids=self.assignees,
                              completion_subscriber_ids=self.completion_subscribers, notify=notify,
                              due_on=self.due_on, starts_on=self.starts_on)

    def __str__(self):
        try:
            return "TodoItem {0.id}: '{0.title}'".format(self)
        except:
            return super(TodoItem, self).__str__()


class Todos(recordings.RecordingEndpoint):
    OBJECT_CLASS = TodoItem
    LIST_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}/todos.json"
    GET_URL = "{base_url}/buckets/{project_id}/todos/{todo_id}.json"
    CREATE_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}/todos.json"
    UPDATE_URL = "{base_url}/buckets/{project_id}/todos/{todo_id}.json"

    COMPLETE_URL = "{base_url}/buckets/{project_id}/todos/{todo_id}/completion.json"
    UNCOMPLETE_URL = "{base_url}/buckets/{project_id}/todos/{todo_id}/completion.json"
    REPOSITION_URL = "{base_url}/buckets/{project_id}/todos/{todo_id}/position.json"

    def list(self, todolist, project=None, status=None, completed=None):
        """
        Retrieve a list of the TodoItem objects in the given TodoList.

        :param todolist: a TodoList object or ID
        :type todolist: basecampy3.endpoints.todolists.TodoList|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param status: set this to "archived" or "trashed" for only TodoItems that match that status
        :type status: str
        :param completed: set to True to only get TodoItems that have been completed, by default only incomplete tasks
                          are listed
        :type completed: bool
        :return: a generator of TodoItem objects in the TodoList specified
        :rtype: collections.Iterable[TodoItem]
        """
        project_id, todolist_id = util.project_or_object(project, todolist)
        params = {}
        if status is not None:
            params['status'] = status
        if completed is not None:
            params['completed'] = six.text_type(completed).lower()  # only accepts 'true' and 'false' not True and False
        url = self.LIST_URL.format(base_url=self.url, todolist_id=todolist_id, project_id=project_id)
        return self._get_list(url, params=params)

    def get(self, todoitem, project=None):
        """
        Get a TodoItem by its ID and project ID.

        :param todoitem: a TodoItem object or ID
        :type todoitem: TodoItem|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :return: a TodoItem
        :rtype: TodoItem
        """
        project_id, todo_id = util.project_or_object(project, todoitem)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, todo_id=todo_id)
        return self._get(url)

    def create(self, content, todolist, project=None, description="", assignee_ids=None,
               completion_subscriber_ids=None, notify=False, due_on=None, starts_on=None):
        """
        Create a new TodoItem in the given TodoList and Project. A TodoList ID and Project ID must be given or just a
        TodoList object. All other parameters are optional.

        :param content: the title or main line of this TodoItem
        :type content: str
        :param todolist: a TodoList object or ID that this TodoItem belongs to
        :type todolist: basecampy3.endpoints.todolists.TodoList|int
        :param project: a Project object or ID that this TodoItem belongs to
        :type project: basecampy3.endpoints.projects.Project|int
        :param description: a longer, HTML-formatted text about this TodoItem
        :type description: str
        :param assignee_ids: a list of Person objects or just their IDs that are responsible for this TodoItem
        :type assignee_ids: list[basecampy3.endpoints.people.Person|int]
        :param completion_subscriber_ids: a list of Person objects or just their IDs that will be notified when this
                                          TodoItem is marked completed
        :type completion_subscriber_ids: list[basecampy3.endpoints.people.Person|int]
        :param notify: set to True if you want assignees to be notified
        :type notify: bool
        :param due_on: the date this item is due. Must be a YYYY-MM-DD string, or a date or datetime object
        :type due_on: str|datetime.datetime|datetime.date
        :param starts_on: the date this item starts. Must be a YYYY-MM-DD string, or a date or datetime object
        :type starts_on: str|datetime.datetime|datetime.date
        :return: the newly created TodoItem
        :rtype: TodoItem
        """
        project_id, todolist_id = util.project_or_object(project, todolist)
        url = self.CREATE_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        data = {
            "content": content,
            "description": description,
            "notify": notify,
        }
        if assignee_ids is not None:
            data['assignee_ids'] = [int(person) for person in assignee_ids]
        if completion_subscriber_ids is not None:
            data['completion_subscriber_ids'] = [int(person) for person in completion_subscriber_ids]
        if due_on is not None:
            data['due_on'] = self._normalize_date(due_on)
        if starts_on is not None:
            data['starts_on'] = self._normalize_date(starts_on)
        return self._create(url, data=data)

    def update(self, todoitem, project=None, content=False, description=False, assignee_ids=False,
               completion_subscriber_ids=False, notify=None, due_on=False, starts_on=False):
        """
        :param todoitem: a TodoItem object or ID
        :type todoitem: TodoItem|int
        :param project: a Project object or ID that this TodoItem belongs to
        :type project: basecampy3.endpoints.projects.Project|int
        :param content: the title or main line of this TodoItem
        :type content: str
        :param description: a longer, HTML-formatted text about this TodoItem
        :type description: str
        :param assignee_ids: a list of Person objects or just their IDs that are responsible for this TodoItem
        :type assignee_ids: list[basecampy3.endpoints.people.Person|int]
        :param completion_subscriber_ids: a list of Person objects or just their IDs that will be notified when this
                                          TodoItem is marked completed
        :type completion_subscriber_ids: list[basecampy3.endpoints.people.Person|int]
        :param notify: set to True if you want assignees to be notified
        :type notify: bool
        :param due_on: the date this item is due. Must be a YYYY-MM-DD string, or a date or datetime object
        :type due_on: str|datetime.datetime|datetime.date
        :param starts_on: the date this item starts. Must be a YYYY-MM-DD string, or a date or datetime object
        :type starts_on: str|datetime.datetime|datetime.date
        """
        project_id, todo_id = util.project_or_object(project, todoitem)
        url = self.UPDATE_URL.format(base_url=self.url, project_id=project_id, todo_id=todo_id)
        if content is False and description is False and assignee_ids is False and completion_subscriber_ids is False \
                and notify is None and due_on is False and starts_on is False:
            raise ValueError("Nothing about this TodoItem would be modified.")
        data = {}
        if content is not False:
            data['content'] = content
        if description is not False:
            data['description'] = description
        if assignee_ids is not False:
            data['assignee_ids'] = assignee_ids
        if completion_subscriber_ids is not False:
            data['completion_subscriber_ids'] = completion_subscriber_ids
        if notify is not None:
            data['notify'] = notify
        if due_on is not False:
            data['due_on'] = self._normalize_date(due_on)
        if starts_on is not False:
            data['starts_on'] = self._normalize_date(starts_on)

        return self._update(url, data)

    def complete(self, todoitem, project=None):
        """
        Mark a TodoItem as complete (check its box).

        :param todoitem: a TodoItem object or ID
        :type todoitem: TodoItem|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        """
        project_id, todo_id = util.project_or_object(project, todoitem)
        url = self.COMPLETE_URL.format(base_url=self.url, project_id=project_id, todo_id=todo_id)

        self._no_response(url, method="POST")

    def uncomplete(self, todoitem, project=None):
        """
        Mark a TodoItem as not complete (un-check its box).

        :param todoitem: a TodoItem object or ID
        :type todoitem: TodoItem|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        """
        project_id, todo_id = util.project_or_object(project, todoitem)
        url = self.UNCOMPLETE_URL.format(base_url=self.url, project_id=project_id, todo_id=todo_id)

        self._no_response(url, method="DELETE")

    def reposition(self, position, todoitem, project=None):
        """
        Moves the TodoItem up or down the list. A position of 1 puts it at the top.

        :param position: the new position for this TodoItem as an integer greater than or equal to 1
        :type position: int
        :param todoitem: a TodoItem object or ID
        :type todoitem: TodoItem|int
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        """
        project_id, todo_id = util.project_or_object(project, todoitem)
        data = {'position': position}
        url = self.REPOSITION_URL.format(base_url=self.url, project_id=project_id, todo_id=todo_id)

        self._no_response(url, data=data)

    @staticmethod
    def _normalize_date(somedate):
        """
        Take a date or datetime and turn them into the string YYYY-MM-DD. If `somedate` is None, returns an
        empty string.

        :param somedate: the date to be coerced into a YYYY-MM-DD format
        :type somedate: datetime.date|datetime.datetime|str
        :return: a string in the format YYYY-MM-DD
        :rtype: str
        """
        if somedate is None:
            return ""

        try:
            somedate = somedate.strftime("%Y-%m-%d")
        except AttributeError:
            pass

        return somedate



from ._base import BasecampObject, BasecampEndpoint


class MessageCategory(BasecampObject):
    def edit(self, project, name=False, icon=False):
        """
        Edit the MessageCategory, changing its name, its icon, or both.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param name: the new name for this category or False for no change
        :type name: str
        :param icon: the new icon for this category or False for no change
        :type icon: str
        """
        self._endpoint.update(project=project, category=self, name=name, icon=icon)

    def delete(self, project):
        self._endpoint.delete(project=project, category=self)

    def __int__(self):
        return int(self.id)

    def __str__(self):
        try:
            return "MessageCategory {0.id}: '{0.name}' {0.icon}".format(self)
        except:
            return super(MessageCategory, self).__str__()


class MessageCategories(BasecampEndpoint):
    OBJECT_CLASS = MessageCategory

    CREATE_URL = "{base_url}/buckets/{project_id}/categories.json"
    DELETE_URL = "{base_url}/buckets/{project_id}/categories/{category_id}.json"
    GET_URL = "{base_url}/buckets/{project_id}/categories/{category_id}.json"
    LIST_URL = "{base_url}/buckets/{project_id}/categories.json"
    UPDATE_URL = "{base_url}/buckets/{project_id}/categories/{category_id}.json"

    def create(self, name, icon, project):
        """

        :param name: the name of this MessageCategory
        :type name: str
        :param icon: the icon for this MessageCategory (usually the unicode character for an emoji)
        :type str: str
        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :return: the newly created MessageCategory object
        :rtype: MessageCategory
        """
        project = int(project)
        data = {
            "name": name,
            "icon": icon,
        }
        url = self.CREATE_URL.format(base_url=self.url, project_id=project)
        new_category = self._create(url, data=data)
        return new_category

    def delete(self, project, category):
        """
        Deletes the MessageCategory from the Project.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param category: a MessageCategory object or ID
        :type category: MessageCategory|int
        """
        project = int(project)
        category = int(category)
        url = self.DELETE_URL.format(base_url=self.url, project_id=project, category_id=category)
        self._delete(url)

    def get(self, project, category):
        """
        Get just one MessageCategory.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param category: a MessageCategory object or ID
        :type category: MessageCategory|int
        :return: a MessageCategory object if found, raises an error if not found
        :rtype: MessageCategory
        """
        project = int(project)
        category = int(category)
        url = self.GET_URL.format(base_url=self.url, project_id=project, category_id=category)
        return self._get(url)

    def list(self, project):
        """
        Get a list of Messages Categories from a given Project.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :return: a list of MessageCategory objects
        :rtype: collections.Iterable[MessageCategory]
        """
        project = int(project)
        url = self.LIST_URL.format(base_url=self.url, project_id=project)
        return self._get_list(url)

    def update(self, project, category, name=False, icon=False):
        """
        Change the name or icon of a given MessageCategory.
        False is a placeholder for no-change. None erases the value like a blank string.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param category: a MessageCategory object or ID
        :type category: MessageCategory|int
        :param name: change the name of the MessageCategory
        :type name: str
        :param icon: change the icon of the MessageCategory
        :type icon: str
        """
        if name is False and icon is False:
            raise ValueError("At least one of subject, content, and category parameters should have a value to "
                             "be changed.")
        project = int(project)
        category = int(category)

        data = {}
        if name is not False:
            data['name'] = "" if name is None else name
        if icon is not False:
            data['icon'] = "" if icon is None else icon

        url = self.UPDATE_URL.format(base_url=self.url, project_id=project, category_id=category)
        self._update(url, data=data)



from ._base import BasecampEndpoint, BasecampObject
from . import projects, templates


class ProjectConstruction(BasecampObject):
    @property
    def project(self):
        """
        When the status of a Project Construction is "completed", we also get the Project data in the JSON response.
        This allows us to construct a Project object without another hit to the API.

        :return: None if the Project isn't ready yet or a Project object if it is
        :rtype: projects.Project
        """
        if 'project' not in self._values:
            return None
        return projects.Project(json_dict=self._values['project'], endpoint=self._endpoint._api.projects)

    @property
    def ready(self):
        """
        A shortcut for self.status == "completed". This variable does not change between invocations of refresh().

        :return: True if and only if this Project Construction's status is "completed"
        :rtype: bool
        """
        return self.status == "completed"

    def __int__(self):
        return int(self.id)


class ProjectConstructions(BasecampEndpoint):
    OBJECT_CLASS = ProjectConstruction

    PROJECT_CREATION_STATUS_URL = "{base_url}/templates/{template_id}/project_constructions/{project_construction_id}.json"
    CREATE_FROM_TEMPLATE_URL = "{base_url}/templates/{template_id}/project_constructions.json"

    def create_project(self, template, name, description=""):
        """
        Creates a new project from the given template. Requires a name and optionally a description.
        Beware that this function does not return a Project object. When made from a template, Projects are made
        asynchronously. This function returns a new `ProjectConstruction` object which allows you to check on the
        status of a project being built from your template.

        :param template: a Template object or ID
        :type template: templates.Template|int
        :param name: a name for this new Project
        :type name: str
        :param description: (optional) a description for this new Project
        :type description: str
        :return: a ProjectConstruction object that can be polled for if the Project is finished being
                 created yet
        :rtype: ProjectConstruction
        """

        template = int(template)

        data = {
            'project': {
                'name': name,
                'description': description,
            }
        }
        url = self.CREATE_FROM_TEMPLATE_URL.format(base_url=self.url, template_id=template)
        project_creation_status = self._create(url, data=data)
        return project_creation_status

    def get(self, template, construction):
        """
        Gets a ProjectConstruction by its ID and the ID of the Template it was created from. This object represents
        whether the Project is finished being created or not.

        :param template: a Template object or ID
        :type template: templates.Template|int
        :param construction: a ProjectConstruction object or ID
        :type construction: ProjectConstruction|int
        :return: a ProjectConstruction object.
        :rtype: ProjectConstruction
        """
        template = int(template)
        construction = int(construction)
        url = self.PROJECT_CREATION_STATUS_URL.format(base_url=self.url, template_id=template,
                                                      project_construction_id=construction)
        return self._get(url)



def project_or_object(project=None, basecamp_object=None, section_name=None):
    """
    The purpose of this function is to provide a Project ID and Object ID given certain kinds of information or to raise
    the appropriate error if the information given isn't enough.

    Given a project object or ID and/or basecamp_object or ID, determines if the ID for both can be found.
    If `project` is an object and `section_name` is given, the ID of the `basecamp_object` can be determined by
    searching the Project object's `dock` list.

    If a `basecamp_object` is given, the Project ID can be found by looking at its "bucket" dictionary.

    If `project` is an int (the Project ID), then `basecamp_object` cannot be None and vice versa. The other variable
    must also be an int or the object.

    :param project:         a Project object or ID. Optional if `basecamp_object` is an Object
    :type project:          basecampy3.endpoints.projects.Project|int

    :param basecamp_object: a Basecamp Object object or ID. Optional if Project is an Object, the BasecampObject to
                                find is a type that appears in a Project's "dock", and `section_name` is specified.
    :type basecamp_object:  basecampy3.endpoints._base.BasecampObject|int

    :param section_name:    the "name" of the section on a project's "dock". Optional if `basecamp_object` is an object
                                or both `project` and `basecamp_object` are IDs
    :type section_name:     str

    :return:                a tuple where the first element is the Project's ID and the second element is the
                                Basecamp Object's ID
    :rtype:                 (int, int)
    """
    if project is None and basecamp_object is None:
        raise ValueError("At least one parameter must be given")
    if project is None:
        project_id = basecamp_object.bucket['id']  # get the Project ID from the Basecamp object
        object_id = int(basecamp_object)
    elif basecamp_object is None:
        if section_name is None:
            raise ValueError("To find an object's ID with only a Project given, you must specify the section's name "
                             "as it appears in the Project's dock. (i.e. todoset, chat, schedule, vault, inbox, "
                             "message_board)")
        for section in project.dock:
            if section['name'] == section_name:
                object_id = section['id']
                break
        else:
            msg = "{project} does not have '{section}' in its dock. Does this project not use {section}?".format(
                project=str(project),
                section=section_name
            )
            ex = AttributeError(msg)
            raise ex
        project_id = int(project)
    else:
        project_id = int(project)
        object_id = int(basecamp_object)

    return project_id, object_id


def normalize_acl(acl):
    """
    Normalize the grant and/or revoke lists we were given. Handle single item, list of items, etc.
    We want to end up with just a list of IDs. Used by Project object to handle modify_access function.

    :param acl: a list of Person objects or just a Person object
    :return: a list of integers representing the IDs of the Person objects in the list given
    """
    try:
        acl = [int(a) for a in acl]  # convert Person objects to list of ID integers
    except TypeError:  # ok maybe we got a single Person instead of a list of them
        acl = [int(acl)]
    return acl



from . import _base, projects, todolists, util


class TodoListGroup(_base.BasecampObject):
    pass


class TodoListGroups(_base.BasecampEndpoint):
    OBJECT_CLASS = TodoListGroup

    LIST_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}/groups.json"
    GET_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_group_id}.json"
    CREATE_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}/groups.json"
    REPOSITION_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_group_id}.json"

    def list(self, todolist, project=None, status=None):
        """
        Retrieves a list of TodoListGroup objects from a given TodoList. Requires either a TodoList object or a
        project ID with a TodoList ID.

        :param todolist: a TodoList object or ID
        :type todolist: todolists.TodoList|int
        :param project: a Project object or ID
        :type project: projects.Project|int
        :param status: (optional) can use "archived" or "trashed" to get only TodoListGroups of that type
        :type status: str
        :return: a list of TodoListGroup objects
        :rtype: collections.Iterable[TodoListGroup]
        """
        params = {}
        if status is not None:
            params['status'] = status
        project_id, todolist_id = util.project_or_object(project, todolist)
        url = self.LIST_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        return self._get_list(url, params=params)

    def get(self, todolist_group, project=None):
        """
        Return a TodoListGroup given a TodoListGroup object or a Project ID and a TodoListGroup ID
        :param todolist_group: a TodoListGroup object or ID
        :type todolist_group: TodoListGroup|int
        :param project:
        :type project: projects.Project|int
        :return:
        """
        project_id, todolist_group_id = util.project_or_object(project, todolist_group)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, todolist_group_id=todolist_group_id)
        return self._get(url)

    def create(self, name, todolist, project=None):
        """
        Create a new TodoListGroup in the given TodoList in the given Project.

        :param name: the name of the new group
        :type name: str
        :param todolist: a TodoList object or ID
        :type todolist: todolists.TodoList|int
        :param project: a Project object or ID
        :type project: projects.Project|int
        :return: a new TodoListGroup object
        :rtype: TodoListGroup
        """
        project_id, todolist_id = util.project_or_object(project, todolist)
        data = {"name": name}
        url = self.CREATE_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        return self._create(url, data=data)

    def reposition(self, position, todolist, project=None):
        """
        Reposition a TodoListGroup in a TodoList.

        :param position: the new position as an integer greater than or equal to 1
        :type position: int
        :param todolist: a TodoList object or ID
        :type todolist: todolists.TodoList|int
        :param project: a Project object or ID
        :type project: projects.Project|int
        """
        project_id, todolist_id = util.project_or_object(project, todolist)
        data = {"position": position}
        url = self.REPOSITION_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        self._update(url, data=data)



from ._base import BasecampEndpoint, BasecampObject


class MessageBoard(BasecampObject):
    """
    A Message Board for posting Messages to.
    """

    def post_message(self, subject, content=None, status="active", category=None):
        """
        Creates a new Message on this MessageBoard.

        :param subject: the subject or title of this Message
        :type subject: str
        :param content: the HTML-formatted body of the message
        :type content: str
        :param status: set to "active" for the message to be live immediately
        :type status: str
        :param category: a MessageCategory object or ID
        :type category: basecampy3.endpoints.message_categories.MessageCategory|int
        :return: the newly created Message object
        :rtype: basecampy3.endpoints.messages.Message
        """
        return self._endpoint._api.messages.create(subject=subject, content=content, status=status,
                                                   category=category, board=self)

    def list(self):
        """
        :return: a list of Message threads on this board.
        :rtype: collections.Iterable[basecampy3.endpoints.messages.Message]
        """
        return self._endpoint._api.messages.list(board=self)

    @property
    def message_types(self):
        return self._endpoint._api.message_categories.list(self.bucket['id'])

    def __int__(self):
        return int(self.id)


class MessageBoards(BasecampEndpoint):
    OBJECT_CLASS = MessageBoard

    GET_URL = "{base_url}/buckets/{project_id}/message_boards/{board_id}.json"

    def get(self, project, board=None):
        """
        Get a single Message Board. If you provide a Project object, you can omit the `board_id` parameter.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param board: a MessageBoard object or ID
        :type board: MessageBoard|int
        :return: a MessageBoard object if found, raises an error if not found
        :rtype: MessageBoard
        """

        project_id = int(project)
        board = int(board) if board else project.message_board.id

        url = self.GET_URL.format(base_url=self.url, project_id=project_id, board_id=board)
        return self._get(url)



import logging
from requests import adapters
from .rated_semaphore import RatedSemaphore
from .dictionary_cache import DictionaryCache
from .constants import RATE_LIMIT_PER_SECONDS, RATE_LIMIT_REQUESTS


class Basecamp3TransportAdapter(adapters.HTTPAdapter):
    """
    Handles API request caching and rate limiting.
    """

    # See https://github.com/basecamp/bc3-api#rate-limiting-429-too-many-requests
    # keeps us under the limit
    SEMAPHORE = RatedSemaphore(RATE_LIMIT_REQUESTS, RATE_LIMIT_PER_SECONDS)

    def __init__(self, cache_backend=None, *args, **kwargs):
        self._cache = DictionaryCache() if cache_backend is None else cache_backend
        super(Basecamp3TransportAdapter, self).__init__(*args, **kwargs)

    def send(self, request, *args, **kwargs):
        """
        Calls the HTTPAdapter.send method with cache headers. Blocks if the request would exceed the rate limit.

        See also:
        https://github.com/basecamp/bc3-api#using-http-caching

        :param request: The :class:`PreparedRequest <PreparedRequest>` being sent.
        :type request: requests.PreparedRequest
        """
        self._set_cache_headers(request)
        method = request.method
        url = request.url
        logging.debug("Consulting with Semaphore")
        with Basecamp3TransportAdapter.SEMAPHORE:  # blocks here until rate limit has cooled off
            logging.debug("OK we can request now.")
            response = super(Basecamp3TransportAdapter, self).send(request, *args, **kwargs)
        if response.status_code == 304:  # not modified; cache hit
            cached_response = self._cache.get_cached_response(method, url)
            return cached_response
        else:
            self._cache_this_response(response)
            return response

    def _cache_this_response(self, response):
        url = response.request.url
        self._cache.set_cached(response)

    def _set_cache_headers(self, request):
        """
        If this request has been made before, get the ETag and Last-Modified headers from the response to it,
        and set the If-None-Match and If-Modified-Since headers on this new request.


        :type request: requests.PreparedRequest
        """
        etag, last_modified = self._cache.get_cached_headers(request.method, request.url)
        if etag:
            request.headers['If-None-Match'] = etag
        if last_modified:
            request.headers['If-Modified-Since'] = last_modified



from . import _base, projects, util


class Recording(_base.BasecampObject):
    """
    Most of the data structures in the Basecamp 3 API are represented as "Recordings", with generic actions
    available to be performed.

    See also:
    https://github.com/basecamp/bc3-api/blob/master/sections/recordings.md
    """
    @property
    def project_id(self):
        """
        :return: the ID of the Project this Recording belongs to.
        :rtype:  int
        """
        return int(self._values['bucket']['id'])

    def archive(self):
        """
        Archive this Recording.
        """
        self._endpoint.archive(project=self.project_id, recording=self)

    def unarchive(self):
        """
        Unarchive this Recording.
        """
        self._endpoint.unarchive(project=self.project_id, recording=self)

    def trash(self):
        """
        Trash this Recording. Trashed items are automatically deleted permanently after 30 days.
        """
        self._endpoint.trash(project=self.project_id, recording=self)


class RecordingEndpoint(_base.BasecampEndpoint):
    OBJECT_CLASS = Recording

    ARCHIVE_URL = "{base_url}/buckets/{project_id}/recordings/{recording_id}/status/archived.json"
    UNARCHIVE_URL = "{base_url}/buckets/{project_id}/recordings/{recording_id}/status/active.json"
    TRASH_URL = "{base_url}/buckets/{project_id}/recordings/{recording_id}/status/trashed.json"

    def archive(self, project=None, recording=None):
        """
        Archive a Recording given its Project and Recording ID.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param recording: a Recording object or ID
        :type recording: Recording|int
        """
        project_id, recording_id = util.project_or_object(project, recording)
        url = self.ARCHIVE_URL.format(base_url=self.url, project_id=project_id, recording_id=recording_id)
        self._no_response(url, method="PUT")

    def unarchive(self, project=None, recording=None):
        """
        Unarchive this Record object.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param recording: a Recording object or ID
        :type recording: Recording|int
        """
        project_id, recording_id = util.project_or_object(project, recording)
        url = self.UNARCHIVE_URL.format(base_url=self.url, project_id=project_id, recording_id=recording_id)
        self._no_response(url, method="PUT")

    def trash(self, project=None, recording=None):
        """
        Trash the given Recording. Trashed items are automatically deleted permanently after 30 days.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project|int
        :param recording: a Recording object or ID
        :type recording: Recording|int
        """
        project_id, recording_id = util.project_or_object(project, recording)
        url = self.TRASH_URL.format(base_url=self.url, project_id=project_id, recording_id=recording_id)
        self._no_response(url, method="PUT")



from . import _base, people, util
from .. import constants
import requests
from ..exc import *
import six
import time


@six.python_2_unicode_compatible
class Project(_base.BasecampObject):
    """
    A Project, also known as a Basecamp, on Basecamp.
    """

    def add_new_user(self, name, email_address, title=None, company_name=None):
        """
        Invite a new user to Basecamp, and add them to this Project. There's no other way to create a new user in
        Basecamp using the API. They must be invited to an initial Project.

        :param name: the user's display name
        :type name: str
        :param email_address: the email address to send the invite to
        :type email_address: str
        :param title: (optional) this user's title at a company
        :type title: str
        :param company_name: (optional) the name of the company this user works at
        :type company_name: str
        :return: the newly created Person object
        :rtype: people.Person
        """
        return self._endpoint.add_new_user(self.id, name=name, email_address=email_address, title=title,
                                           company_name=company_name)

    def modify_access(self, grant=None, revoke=None):
        """
        Change who is allowed to access this Project. `grant` and `revoke` take a list of Person objects or a list of
        Person IDs as integers. They can both be specified at the same time.

        :param grant: a list of Person objects or Person IDs (ints) to grant access
        :type grant: list[people.Person|int]
        :param revoke: a list of Person objects or Person IDs (ints) to remove from the Project
        :type revoke: list[people.Person|int]
        :return: a response object
        :rtype: requests.Response
        """
        return self._endpoint.modify_access(self.id, grant=grant, revoke=revoke)

    def trash(self):
        """
        Trash this project. Trashed projects will be deleted from Basecamp 3 after 30 days.
        """
        self._endpoint.trash(self.id)

    def update(self, name=False, description=False):
        """
        Change this Project's name and/or description.

        :param name: the new name for this Project. A value of False means no change.
        :type name: str
        :param description: the new description for this Project. A value of False means no change.
        :type description: str
        """
        self._endpoint.update(self.id, name=name, description=description)

    @property
    def campfire(self):
        """
        :return: the Campfire object associated with this Project.
        :rtype: basecampy3.endpoints.campfires.Campfire
        """
        section = self._get_dock_section(constants.DOCK_NAME_CAMPFIRE)
        return self._endpoint._api.campfires.get(campfire=section['id'], project=self.id)

    @property
    def message_board(self):
        """
        :return: the MessageBoard object associated with this Project
        :rtype: basecampy3.endpoints.message_boards.MessageBoard
        """
        section = self._get_dock_section(constants.DOCK_NAME_MESSAGE_BOARD)
        return self._endpoint._api.message_boards.get(board=section['id'], project=self.id)

    @property
    def todoset(self):
        """
        :return: the TodoSet object associated with this Project
        :rtype: basecampy3.endpoints.todosets.TodoSet
        """
        section = self._get_dock_section(constants.DOCK_NAME_TODOS)
        return self._endpoint._api.todosets.get(todoset=section['id'], project=self.id)

    @property
    def people(self):
        """
        A list of people who currently have access to this Project.

        :return: a list of Person objects
        :rtype: collections.Iterable[basecampy3.endpoints.people.Person]
        """
        return self._endpoint._api.people.list(project=self.id)

    def _get_dock_section(self, name):
        for section in self.dock:
            if section['name'] == name:
                return section

    def __str__(self):
        try:
            return u"Project {0.id}: '{0.name}'".format(self)
        except:
            return super(Project, self).__str__()

    def __repr__(self):
        repr_dict = {k: v for k, v in self._values.items() if k not in ("dock",)}
        return "Project(%s)" % repr(repr_dict)


class Projects(_base.BasecampEndpoint):
    OBJECT_CLASS = Project

    CREATE_URL = "{base_url}/projects.json"
    GET_URL = "{base_url}/projects/{project_id}.json"
    LIST_URL = "{base_url}/projects.json"
    MODIFY_ACCESS_URL = "{base_url}/projects/{project_id}/people/users.json"
    TRASH_URL = "{base_url}/projects/{project_id}.json"
    UPDATE_URL = "{base_url}/projects/{project_id}.json"

    CREATION_FROM_TEMPLATE_TIMEOUT = 10

    def add_new_user(self, project, name, email_address, title=None, company_name=None):
        """
        Invite a new user to Basecamp, and add them to the given Project. There's no other way to create a new user in
        Basecamp using the API. They must be invited to an initial Project.

        :param project: a Project object or ID
        :type project: Project|int
        :param name: the user's display name
        :type name: str
        :param email_address: the email address to send the invite to
        :type email_address: str
        :param title: (optional) this user's title at a company
        :type title: str
        :param company_name: (optional) the name of the company this user works at
        :type company_name: str
        :return: the newly created Person object
        :rtype: basecampy3.endpoints.people.Person
        """
        create = {
            'name': name,
            'email_address': email_address,
        }
        if title is not None:
            create['title'] = title
        if company_name is not None:
            create['company_name'] = company_name
        data = {
            'create': [
                create,
            ]
        }

        project = int(project)
        url = self.MODIFY_ACCESS_URL.format(base_url=self.url, project_id=project)
        resp = self._api._session.put(url, json=data)
        result = resp.json()
        new_user = people.Person(json_dict=result['granted'][0], endpoint=self._api.people)
        return new_user

    def create(self, name, description="", template=None):
        """
        Create a new Project on Basecamp

        :param name: the name for this new Project
        :type name: str
        :param description: (optional) a description for this new Project
        :type description: str
        :param template: a Template object or ID
        :type template:  basecampy3.endpoints.templates.Template|int
        :return: the newly created Project object
        :rtype: Project
        """
        if template is not None:
            return self._create_from_template(name, description, template)
        data = {
            'name': name,
            'description': description,
        }
        url = self.CREATE_URL.format(base_url=self.url)
        new_project = self._create(url, data=data)
        return new_project

    def get(self, project):
        """
        Get just one Project.

        :param project: a Project object or ID
        :type project: Project|int
        :return: a Project object if found, raises an error if not found
        :rtype: Project
        """
        project = int(project)
        url = self.GET_URL.format(base_url=self.url, project_id=project)
        return self._get(url)

    def list(self, status=None):
        """
        Get a list of Basecamp projects visible to the user.

        :param status: optionally can be 'archived' or 'trashed' to get projects of that type
        :type status: str
        :return: a generator of Project objects
        :rtype: collections.Iterable[Project]
        """
        params = {}
        if status is not None:
            params['status'] = status
        url = self.LIST_URL.format(base_url=self.url)
        return self._get_list(url, params)

    def modify_access(self, project, grant=None, revoke=None):
        """
        Change who is allowed to access a given Project. `grant` and `revoke` take a list of Person objects or a list
        of Person IDs as integers. They can both be specified at the same time.

        :param project: a Project object or ID
        :type project: Project|int
        :param grant: a list of Person objects or Person IDs (ints) to grant access
        :type grant: list[people.Person|int]
        :param revoke: a list of Person objects or Person IDs (ints) to remove from the Project
        :type revoke: list[people.Person|int]
        :return: a response object
        :rtype: requests.Response
        """
        if grant is None and revoke is None:
            raise ValueError("Can't modify access without specifying at least the grant or revoke parameters")
        data = {}
        if grant:
            data['grant'] = util.normalize_acl(grant)
        if revoke:
            data['revoke'] = util.normalize_acl(revoke)
        project = int(project)
        url = self.MODIFY_ACCESS_URL.format(base_url=self.url, project_id=project)
        resp = self._api._session.put(url, json=data)
        return resp

    def trash(self, project):
        """
        Trash the given project. Trashed projects will be deleted from Basecamp 3 after 30 days.
        :param project: a Project object or ID
        """
        project = int(project)
        url = self.TRASH_URL.format(base_url=self.url, project_id=project)
        self._delete(url)

    def update(self, project, name=False, description=False):
        """
        Change the given Project's name and/or description.

        :param project: a Project object or ID
        :type project: Project|int
        :param name: the new name for this Project. A value of False means no change.
        :type name: str
        :param description: the new description for this Project. A value of False means no change.
        :type description: str

        """
        if name is False and description is False:
            raise ValueError("Nothing to update for Project %s" % project)
        project = int(project)
        data = {}
        if name is not False:
            data['name'] = name
        if description is not False:
            data['description'] = description
        url = self.UPDATE_URL.format(base_url=self.url, project_id=project)
        return self._update(url, data=data)

    def _create_from_template(self, name, description, template, timeout=CREATION_FROM_TEMPLATE_TIMEOUT):
        """
        Synchronously creates a Project from a Template, handling the polling of a ProjectConstruction object so that
        the user doesn't have to.

        If a Project is not finished being created before the end of `timeout`, ProjectCreationTimedOutError is raised.

        :param name: the name of the new Project
        :type name: str
        :param description: (optional) the description of the new Project
        :type description: str
        :param template: a Template object or ID
        :type template: basecampy3.endpoints.templates.Template|int
        :param timeout: number of seconds to wait. If not specified, a default timeout defined in the class will be used
        :type timeout: int
        :return: a new Project created from the given Template
        :rtype: Project
        """
        creation_status = self._api.project_constructions.create_project(template=template, name=name,
                                                                         description=description)
        counter = 0
        # usually projects get created pretty fast. Pause for 1 second before checking again.
        while not creation_status.ready:
            time.sleep(1)
            creation_status.refresh()
            counter += 1
            if counter >= timeout:
                msg = u"Project took more than " \
                      u"{seconds} seconds to be created.".format(seconds=timeout)
                ex = ProjectCreationTimedOutError(message=msg)
                raise ex
        return creation_status.project



from ._base import BasecampEndpoint, BasecampObject
from ..constants import DOCK_NAME_CAMPFIRE
from . import util


class CampfireLine(BasecampObject):
    """
    A single line of a Campfire. Basically, a single chat message.
    """
    def __int__(self):
        return int(self.id)

    def __str__(self):
        try:
            return "[{}] {}: '{}'".format(self.created_at, self.creator['name'], self.content)
        except:
            return super(CampfireLine, self).__str__()


class CampfireLines(BasecampEndpoint):
    OBJECT_CLASS = CampfireLine

    LIST_URL = "{base_url}/buckets/{project_id}/chats/{campfire_id}/lines.json"
    GET_URL = "{base_url}/buckets/{project_id}/chats/{campfire_id}/lines/{campfire_line_id}.json"
    CREATE_URL = "{base_url}/buckets/{project_id}/chats/{campfire_id}/lines.json"
    DELETE_URL = "{base_url}/buckets/{project_id}/chats/{campfire_id}/lines/{campfire_line_id}.json"

    def list(self, project=None, campfire=None):
        """
        Get a list of chat lines in the given Campfire.
        If `campfire` is a Campfire object, the `project` parameter can be omitted.
        If `project` is a Project object, the `campfire` parameter can be omitted.

        :param project: a Project object or Project ID
        :param campfire: a Campfire object or Campfire ID
        :return: a list of messages in the project
        """
        project_id, campfire_id = util.project_or_object(project, campfire, section_name=DOCK_NAME_CAMPFIRE)
        url = self.LIST_URL.format(base_url=self.url, project_id=project_id, campfire_id=campfire_id)
        return self._get_list(url)

    def get(self, campfire_line, project=None, campfire=None):
        """
        Get a single Campfire Line

        :param campfire_line: Campfire Line ID
        :param project: a Project object or Project ID
        :param campfire: a Campfire object or Campfire ID
        :return: a Campfire Line object
        """
        campfire_line = int(campfire_line)
        project_id, campfire_id = util.project_or_object(project, campfire, section_name=DOCK_NAME_CAMPFIRE)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, campfire_id=campfire_id,
                                  campfire_line_id=campfire_line)
        return self._get(url)

    def create(self, content, project=None, campfire=None):
        """
        Post a new line (message) in the given Campfire from the current user with the provided `content` string.
        Can be formatted in HTML.

        :param content: message string that can be HTML-formatted
        :param project: a Project object or Project ID
        :param campfire: a Campfire object or Campfire ID
        :return: the newly created Campfire Line
        """
        project_id, campfire_id = util.project_or_object(project, campfire, section_name=DOCK_NAME_CAMPFIRE)
        data = {
            'content': content,
        }
        url = self.CREATE_URL.format(base_url=self.url, project_id=project_id, campfire_id=campfire_id)
        new_line = self._create(url, data=data)
        return new_line

    def delete(self, campfire_line, project=None, campfire=None):
        """
        Delete a line from the campfire.

        :param campfire_line: the Campfire Line object or ID
        :param project: a Project object or Project ID
        :param campfire: a Campfire object or Campfire ID
        """
        campfire_line = int(campfire_line)
        project_id, campfire_id = util.project_or_object(project, campfire, section_name=DOCK_NAME_CAMPFIRE)
        url = self.DELETE_URL.format(base_url=self.url, project_id=project_id, campfire_id=campfire_id,
                                     campfire_line_id=campfire_line)
        self._delete(url)



from threading import Timer
try:
    from threading import _Semaphore as Semaphore
except ImportError:
    from threading import Semaphore
import time


class RatedSemaphore(Semaphore):
    """
    Limit to 1 request per `period / value` seconds (over long run). Used to put a limit on time-restricted resources.

    For instance, if we are allowed only 50 calls in 10 seconds to an API, you can use:

    ```
    semaphore = RatedSemaphore(50, 10)
    with semaphore:
        call_a_thing()
    ```

    This means we can call as fast as we like and only start blocking when we would exceed our limit.

    Copied with some modifications from:
    https://stackoverflow.com/a/16686329/489667
    """
    def __init__(self, value=1, period=1):
        """
        :param value: the number of tokens in a given period. This replenishes over time but blocks when it hits 0.
        :type value: int
        :param period: the time, in seconds, in a period
        :type period: int
        """
        super(RatedSemaphore, self).__init__(value)
        t = Timer(period, self._add_token_loop, kwargs={'time_delta': float(period) / value})
        t.daemon = True
        t.start()
        self.t = t

    def _add_token_loop(self, time_delta):
        """
        Add token every time_delta seconds.

        :param time_delta: the time between adding a token
        :type time_delta: float
        """
        while True:
            try:
                super(RatedSemaphore, self).release()
            except ValueError:  # ignore if already max possible value
                pass
            time.sleep(time_delta)  # ignore EINTR

    def release(self):
        """
        Not allowed. Only the internal timer can call release.
        """
        pass  # called by the `with` statement so just ignore it



import six


@six.python_2_unicode_compatible
class Basecamp3Error(Exception):
    def __init__(self, response=None, message=None):
        super(Basecamp3Error, self).__init__()
        self.response = response
        self.message = message

    def __str__(self):
        if self.response is not None:
            return u"%s %s %s" % (self.response.status_code, self.response.reason, self.response.text)
        if self.message is not None:
            return self.message


class InvalidRefreshTokenError(Basecamp3Error):
    pass


class InvalidUserCodeError(Basecamp3Error):
    pass


class UnauthorizedError(Basecamp3Error):
    pass


class UnknownAccountIDError(Basecamp3Error):
    pass


class ProjectCreationTimedOutError(Basecamp3Error):
    pass


class NoDefaultConfigurationFound(Basecamp3Error):
    pass



from . import _base
import six


@six.python_2_unicode_compatible
class Person(_base.BasecampObject):
    """
    A user profile on Basecamp 3
    """

    def __str__(self):
        try:
            return u"Person {0.id}: '{0.name}'".format(self)
        except:
            return super(Person, self).__str__()


class People(_base.BasecampEndpoint):
    OBJECT_CLASS = Person

    LIST_PEOPLE_URL = "{base_url}/people.json"
    LIST_BY_PROJECT_URL = "{base_url}/projects/{project_id}/people.json"
    LIST_PINGABLE_URL = "{base_url}/circles/people.json"
    GET_PERSON_URL = "{base_url}/people/{person_id}.json"
    GET_MYSELF_URL = "{base_url}/my/profile.json"

    def list(self, project=None):
        """
        Get a list of people visible to the user.

        :param project: optionally can pick a project to list the people who are members of it
        :type project: basecampy3.endpoints.projects.Project|int
        :return: a list of Person objects
        :rtype: collections.Iterable[Person]
        """
        if project is not None:
            project = int(project)
        if project is not None:
            url = self.LIST_BY_PROJECT_URL.format(base_url=self.url, project_id=project)
        else:
            url = self.LIST_PEOPLE_URL.format(base_url=self.url)
        return self._get_list(url)

    def list_pingable(self):
        """
        Get a list of people that the user can ping
        :return: a list of Person objects
        :rtype: collections.Iterable[Person]
        """
        url = self.LIST_PINGABLE_URL.format(base_url=self.url)
        return self._get_list(url)

    def get(self, oid=None):
        """
        Get Person by ID. If no ID is given, returns the current user's Person
        :param oid: the person's ID
        :type oid: Person|int
        :return: a Person object
        :rtype: Person
        """
        if oid is None:
            url = self.GET_MYSELF_URL.format(base_url=self.url)
        else:
            oid = int(oid)
            url = self.GET_PERSON_URL.format(base_url=self.url, person_id=oid)
        return self._get(url)



from ._base import BasecampEndpoint, BasecampObject
import six


@six.python_2_unicode_compatible
class Template(BasecampObject):
    """
    A Project Template or Blueprint from which you can make Projects that have the same start.
    """

    def create_project(self, name, description=""):
        return self._endpoint._api.project_constructions.create_project(self, name, description)

    def trash(self):
        return self._endpoint.trash(self)

    def __int__(self):
        return int(self.id)

    def __str__(self):
        try:
            return u"Template {0.id}: '{0.name}'".format(self)
        except:
            return super(Template, self).__str__()

    def __repr__(self):
        repr_dict = {k: v for k, v in self._values.items() if k not in ("dock",)}
        return u"Template(%s)" % repr(repr_dict)


class Templates(BasecampEndpoint):
    OBJECT_CLASS = Template

    CREATE_URL = "{base_url}/templates.json"
    GET_URL = "{base_url}/templates/{template_id}.json"
    LIST_URL = "{base_url}/templates.json"
    UPDATE_URL = "{base_url}/templates/{template_id}.json"
    TRASH_URL = "{base_url}/templates/{template_id}.json"

    def create(self, name, description=""):
        data = {
            'name': name,
            'description': description,
        }
        url = self.CREATE_URL.format(base_url=self.url)
        new_template = self._create(url, data=data)
        return new_template

    def get(self, template):
        """
        Get just one Template.

        :param template: a Template object or ID
        :return: a Template object if found, raises an error if not found
        """
        template = int(template)
        url = self.GET_URL.format(base_url=self.url, template_id=template)
        return self._get(url)

    def list(self, status=None):
        """
        Get a list of Basecamp Project Templates visible to the user.

        :param status: optionally can be 'archived' or 'trashed' to get templates of that type
        :return: a list of Template objects
        """
        params = {}
        if status is not None:
            params['status'] = status
        url = self.LIST_URL.format(base_url=self.url)
        return self._get_list(url, params)

    def update(self, template, name=False, description=False):
        """
        Change the given Template's name and/or description.

        :param template: a Template object or ID
        :param name: the new name for this Template. A value of False means no change.
        :param description: the new description for this Template. A value of False means no change.
        """
        if name is False and description is False:
            raise ValueError("Nothing to update for Template %s" % template)
        template = int(template)
        data = {}
        if name is not False:
            data['name'] = name
        if description is not False:
            data['description'] = description
        url = self.UPDATE_URL.format(base_url=self.url, template_id=template)
        self._update(url, data=data)

    def trash(self, template):
        template = int(template)
        url = self.TRASH_URL.format(base_url=self.url, template_id=template)
        self._delete(url)






from setuptools import setup

VERSION = "0.1.5"


with open('README.md', 'r') as infile:
    long_description = infile.read()

setup(
    name='basecampy3',
    version=VERSION,
    packages=[
        'basecampy3',
        'basecampy3.endpoints'
    ],
    install_requires=[
        "beautifulsoup4==4.6.0",
        "certifi>=2018.1.18",
        "chardet==3.0.4",
        "idna==2.6",
        "python-dateutil==2.7.2",
        "requests==2.18.4",
        "six==1.11.0",
        "urllib3==1.22",
    ],
    entry_points={
        'console_scripts': [
            'bc3 = basecampy3.bc3_cli:main',
        ],
    },
    url='https://github.com/phistrom/basecampy3',
    license='MIT',
    author='Phillip Stromberg',
    author_email='phillip@4stromberg.com',
    description='Aims to be the easiest to use version of the Basecamp 3 API',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ),
    long_description=long_description,
    long_description_content_type="text/markdown"
)



from collections import OrderedDict
from .response_cache import ResponseCache


class DictionaryCache(ResponseCache):
    """
    A simple cache used by default. Not good for multi-threaded or multi-host instances where you might need to cache
    your requests in something like Redis.
    """
    def __init__(self, max_entries=20):
        """
        :type max_entries: int
        """
        super(DictionaryCache, self).__init__()
        self.__max_entries = 0
        self.max_entries = max_entries
        self._cache_dict = OrderedDict()

    @property
    def max_entries(self):
        return self.__max_entries

    @max_entries.setter
    def max_entries(self, value):
        value = int(value)
        if value < 0:
            raise ValueError("max_entries cannot be negative")
        self.__max_entries = value

    def get_cached_headers(self, method, url):
        try:
            key = (method, url)
            etag, last_modified, _ = self._cache_dict[key]
            return etag, last_modified
        except KeyError:
            return None, None

    def get_cached_response(self, method, url):
        item = self._cache_dict[(method, url)]
        response = item[2]  # our cached items are a tuple as etag, last_modified, and response
        return response

    def set_cached(self, response):
        key = (response.request.method, response.request.url)
        etag = response.headers.get('ETag')
        last_modified = response.headers.get('Last-Modified')
        self._add_to_cache(key, etag, last_modified, response)

    def _add_to_cache(self, key, etag, last_modified, response):
        item = (etag, last_modified, response)

        try:
            del self._cache_dict[key]  # pop this response out of the cache if it's in there already
        except KeyError:
            pass

        while len(self._cache_dict) >= self.max_entries:  # pop off oldest entries until within limit
            self._cache_dict.popitem(last=False)

        # it's now the freshest item in the cache
        self._cache_dict[key] = item



from . import recordings, todosets, util
from .. import constants


class TodoList(recordings.Recording):
    @property
    def todoset_id(self):
        """
        :return: the ID of the TodoSet that this TodoList belongs to
        :rtype: int
        """
        return int(self._values['parent']['id'])

    def create(self, content, description="", assignee_ids=None, completion_subscriber_ids=None, notify=False,
               due_on=None, starts_on=None):
        """
        Create a new TodoItem in the given TodoList and Project. A TodoList ID and Project ID must be given or just a
        TodoList object. All other parameters are optional.

        :param content: the title or main line of this TodoItem
        :type content: str
        :param description: a longer, HTML-formatted text about this TodoItem
        :type description: str
        :param assignee_ids: a list of Person objects or just their IDs that are responsible for this TodoItem
        :type assignee_ids: list[basecampy3.endpoints.people.Person|int]
        :param completion_subscriber_ids: a list of Person objects or just their IDs that will be notified when this
                                          TodoItem is marked completed
        :type completion_subscriber_ids: list[basecampy3.endpoints.people.Person|int]
        :param notify: set to True if you want assignees to be notified
        :type notify: bool
        :param due_on: the date this item is due. Must be a YYYY-MM-DD string, or a date or datetime object
        :type due_on: str|datetime.datetime|datetime.date
        :param starts_on: the date this item starts. Must be a YYYY-MM-DD string, or a date or datetime object
        :type starts_on: str|datetime.datetime|datetime.date
        :return: the newly created TodoItem
        :rtype: basecampy3.endpoints.todos.TodoItem
        """
        return self._endpoint._api.todos.create(content=content, description=description, assignee_ids=assignee_ids,
                                                completion_subscriber_ids=completion_subscriber_ids, notify=notify,
                                                due_on=due_on, starts_on=starts_on, todolist=self,
                                                project=self.project_id)

    def create_group(self, name):
        """
        Create a new TodoListGroup in this TodoList.

        :param name:
        :type name: str
        :return: the newly created TodoListGroup object
        :rtype: basecampy3.endpoints.todolist_groups.TodoListGroup
        """
        return self._endpoint._api.todolist_groups.create(name=name, todolist=self, project=self.project_id)


    def list(self, status=None, completed=None):
        return self._endpoint._api.todos.list(todolist=self, project=self.project_id,
                                              status=status, completed=completed)

    def list_groups(self, status=None):
        return self._endpoint._api.todolist_groups.list(todolist=self, project=self.project_id, status=status)

    def __str__(self):
        try:
            return "TodoList {0.id}: '{0.name}'".format(self)
        except:
            return super(TodoList, self).__str__()


class TodoLists(recordings.RecordingEndpoint):
    OBJECT_CLASS = TodoList

    LIST_URL = "{base_url}/buckets/{project_id}/todosets/{todoset_id}/todolists.json"
    GET_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}.json"
    CREATE_URL = "{base_url}/buckets/{project_id}/todosets/{todoset_id}/todolists.json"
    UPDATE_URL = "{base_url}/buckets/{project_id}/todolists/{todolist_id}.json"

    def list(self, project=None, todoset=None, status=None):
        """
        Get a list of TodoList objects. Requires either a Project object or a project ID with a TodoSet ID.

        :param project: a Project object or ID
        :type project: basecampy3.endpoints.projects.Project
        :param todoset: a TodoSet object or ID
        :type todoset: todosets.TodoSet
        :param status: (optional) get TodoLists that are "archived" or "trashed"
        :type status: str
        :return: a list of TodoList objects from the given Project
        :rtype: collections.Iterable[TodoList]
        """
        params = {}
        if status is not None:
            params['status'] = status
        project_id, todoset_id = util.project_or_object(project, todoset, section_name=constants.DOCK_NAME_TODOS)
        url = self.LIST_URL.format(base_url=self.url, project_id=project_id, todoset_id=todoset_id)
        return self._get_list(url, params=params)

    def get(self, todolist, project=None):
        """
        Get a TodoList by its ID and a Project's ID or just a TodoList object.

        :param todolist: a TodoList object or ID
        :type todolist: TodoList|int
        :param project: a Project object or ID (optional if `todolist` is an object)
        :type project: basecampy3.endpoints.projects.Project|int
        :return: a TodoList object
        :rtype: TodoList
        """
        project_id, todolist_id = util.project_or_object(project, todolist)
        url = self.GET_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        return self._get(url)

    def create(self, name, description=None, project=None, todoset=None):
        """
        Create a new TodoList in the given Project.

        :param name: the name for this TodoList
        :type name: str
        :param description:
        :type description: str
        :param project:
        :type project: basecampy3.endpoints.projects.Project|int
        :param todoset:
        :type todoset: todosets.TodoSet|int
        :return: a new TodoList object
        :rtype: TodoList
        """
        project_id, todoset_id = util.project_or_object(project, todoset, section_name=constants.DOCK_NAME_TODOS)
        data = {
            "name": name,
            "description": description
        }
        url = self.CREATE_URL.format(base_url=self.url, project_id=project_id, todoset_id=todoset_id)
        return self._create(url, data=data)

    def update(self, todolist, project=None, name=False, description=False):
        if name is False and description is False:
            raise ValueError("Nothing to update for TodoList %s" % todolist)
        project_id, todolist_id = util.project_or_object(project, todolist)
        data = {}
        if name is not False:
            data['name'] = name
        if description is not False:
            data['description'] = description
        url = self.UPDATE_URL.format(base_url=self.url, project_id=project_id, todolist_id=todolist_id)
        return self._update(url, data=data)



import abc
import six


@six.add_metaclass(abc.ABCMeta)
class ResponseCache(object):
    """
    Extend this class to create your own caching for Basecamp 3 API calls. The responsibility of the cache is to
    remember the ETag and Last-Modified headers returned by a call and send it on a subsequent call. Basecamp 3 will
    respond with a "304 Not Modified" and no content if the response would be the same as last time. Then you can just
    return the cached Response.

    requests.Response objects can be Pickled, so they can be stored externally like in Redis, a file, or a database.
    """

    @abc.abstractmethod
    def get_cached_headers(self, method, url):
        """
        Given a request, gets the cached headers if we've seen this exact request METHOD and URL before.
        The headers will be given in the form of tuple(ETAG, LAST_MODIFIED).
        :param method: the HTTP method in all caps (i.e. 'GET', 'POST', 'PUT')
        :type method: str
        :param url: the URL of the request
        :type url: str
        :return: the headers as a 2-element tuple. One or both can be `None` if the headers have not been cached before.
        :rtype: tuple(str)
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_cached_response(self, method, url):
        """
        Get a requests.Response object that was cached. The key is the METHOD and URL.

        :param method: the HTTP method in all caps (i.e. 'GET', 'POST', 'PUT')
        :type method: str
        :param url: the URL of the request
        :type url: str
        :return: the Response object that has been cached since the last time this endpoint was called
        :rtype: requests.Response
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_cached(self, response):
        """
        Cache this Response object.

        :param response: the Response object from a successful call to be retrieved later
        :type response: requests.Response
        """
        raise NotImplementedError()
